﻿using System;
using System.Collections.Generic;
using Produmex.Foundation.Collections;
using System.Collections.ObjectModel;
using System.Linq;
using System.Data;
using System.Globalization;
using System.Reflection;
using System.Text;
using Produmex.Foundation.Barcode;
using Produmex.Foundation.Convertors;
using Produmex.Foundation.Data.DbClient;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions.Tables;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.Messages;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.SlimScreen.Interfaces;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions.KnownDataSets;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.DataObjects;
using Produmex.Sbo.Logex.Quorum.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.Providers;
using Produmex.Foundation.Data.Sbo;
using Produmex.Sbo.Logex.Data.Generators.LocationController;
using Produmex.Foundation.Data.Sbo.Generators.BarcodeController;
using Produmex.Sbo.Logex.Quorum.Providers;
using Produmex.Sbo.Logex.Quorum.BusinessObjects;
using Produmex.Foundation;
using Produmex.Foundation.Utilities;
using Produmex.Foundation.Globalization;


// PmxWorkflowExecutionTypes.HOOK_FLOW
namespace Produmex.Sbo.Logex.WorkflowScripts
{
    public class WorkflowScript_AfterItemPickedHookScript : WorkflowInstanceScriptBase
    {
        private static class Definitions
        {
            public static string BLADE = "B";
            public static int ZERO = 0;
            public const int MIMIMUM_NUMBER_OF_CHARACTERS_FOR_REASON = 10;
        }

        private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);

        // Input parameters * do not change *
        public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
        public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;

        public ReadOnlyBinder<Collection<PickingItemInfo>> PickedItemInfos;
        public ReadOnlyBinder<int?> LUID;
        public ReadOnlyBinder<Collection<string>> ListOfSerialNumbers;
        public ReadOnlyBinder<int> DocEntry;
        public ReadOnlyBinder<string> ObjType;

        public WorkflowScript_AfterItemPickedHookScript(WorkflowInstanceBase parent, WorkflowInstanceFactory factory)
            : base(parent, factory)
        {
        }

        #region WorkflowInstanceScriptBase Members

        protected override void Execute()
        {
            // Parameters in scope
            Session session = GetScopeParameter("Session") as Session;
            ISboProviderService sboProviderService = GetScopeParameter("<WwfService>ISboService") as ISboProviderService;

            IBarcodeController barcodeController = null;
            WorkflowInstanceBase m_enterQuantitySubFlow;
            m_enterQuantitySubFlow = this.Create("EnterQuantityScript");

            sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
                delegate (PmxDbConnection conn, object[] parameters)
                {
                    PmxBarcodeProvider bcProv = new PmxBarcodeProvider(conn);
                    barcodeController = bcProv.GetBarcodeController();

                    return null;
                });

            WorkflowHelper methodHelper = new WorkflowHelper()
            {
                PmxOseCompany = PmxOseCompany,
                CultureInfo = this.DefaultCultureInfo,
                ShowScreen = session.ShowScreen,
                DeviceInstanceId = session.DeviceInstanceId,
                WaitForMessage = this.WaitForMessage,
                StartSubFlowAndWait = this.StartSubFlowAndWait,
                Create = this.Create,
                SboProviderService = sboProviderService,
                BarcodeController = barcodeController,
                ResourceStore = base.ResourceStore,
                WhsInfo = sboProviderService.InvokeMethodWithDbConnection<PmxOseWarehouseInfo>(false, false, null, null, new DelegateWithPmxDbConnection<PmxOseWarehouseInfo>(ExecuteGetPmxWarehouseCode), session.DeviceInstanceId)

            };


            BladeSwapDataCheckResult bladeSwapData = CheckAndGatherDataForProcessing(methodHelper, PickedItemInfos, LUID, ListOfSerialNumbers, DocEntry, ObjType);

            if (bladeSwapData.Step == Steps.Stop || bladeSwapData.Step == Steps.Return)
            {
                return;
            }

        Step_ShowSwapInstructions:

            bool isThereBladSwapInstruction = String.IsNullOrWhiteSpace(bladeSwapData.FanSwapInstructions);

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                DefaultCultureInfo.Get(), BuildParamCollection(
                    "MessageKey", isThereBladSwapInstruction ? "MSG_BLADE_SWAP_INSTRUCTION_IS_MISSING" : bladeSwapData.FanSwapInstructions,
                    "NoTranslationOfMessageKey", !isThereBladSwapInstruction,
                    "IsError", false,
                    "ShowButton", true));
            WaitForMessage();

        Step_GetQuantity:

            Steps screenResult = GetBladeQuantity(methodHelper, bladeSwapData, barcodeController);

            if (screenResult == Steps.Return)
            {
                goto Step_ShowSwapInstructions;
            }

            List<MoveItemInfoExtended> moveItems = GetPickedItemsAsMoveItems(PickedItemInfos);

            LocationResult locResult = GetDestinationLocation(methodHelper, moveItems, false, bladeSwapData.BladeItemCodeToSwap);

            if (locResult.Step == Steps.Next)
            {
                DoDeliveryStockCreationBooking(methodHelper, bladeSwapData, locResult);
            }
            else if (locResult.Step == Steps.Return)
            {
                goto Step_GetQuantity;
            }

			PrintSSCCLabel(methodHelper, locResult.NeedSSCC, locResult.DestinationLocationCode, methodHelper.DeviceInstanceId, locResult.LUID);

			PrintItemLabel(methodHelper, bladeSwapData, locResult);

		}

		private Steps GetBladeQuantity(WorkflowHelper methodHelper, BladeSwapDataCheckResult bladeSwapData, IBarcodeController barcodeController)
		{
			methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IEnterQuantityWithItemInfoScreen),
							DefaultCultureInfo.Get(), BuildParamCollection(
								"InitialErrorKey", methodHelper.InitialErrorKey,
								"TitleKey", "MSG_TITLE_ENTER_BLADE_QUANTITY",
								"BarcodeController", barcodeController,
								"NumberOfDigitsAfterDecimal", 0,
								"InitialQuantity", bladeSwapData.BladeQuantityToSwap,
								"MaximumQuantity", -1,
								"UOM", null,
								"ItemDescription", bladeSwapData.BladeItemDescriptionToSwap,
								"ItemCode", bladeSwapData.BladeItemCodeToSwap,
								"ItemBarcode", null,
								"BatchNumber", null,
								"BestBeforeDate", null,
								"StillToQuantity", bladeSwapData.BladeQuantityToSwap
							   ));

			methodHelper.Msg = WaitForMessage();

			if (methodHelper.Msg.Name.EndsWith(".QuantityEntered"))
			{
				bladeSwapData.BladeQuantityToSwap = ExtractParameter<double>(methodHelper.Msg.Parameters, "quantity");
				return Steps.Next;

			}
			else if (methodHelper.Msg.Name.EndsWith(".BackRequested"))
			{

				return Steps.Return;
			}

			return Steps.Next;

		}

		private List<MoveItemInfoExtended> GetPickedItemsAsMoveItems(Collection<PickingItemInfo> pickedItemInfos)
		{
			List<MoveItemInfoExtended> result = new List<MoveItemInfoExtended>();
			foreach (PickingItemInfo item in pickedItemInfos)
			{
				MoveItemInfoExtended moveItem = new MoveItemInfoExtended();
				moveItem.ItemCode = item.ItemCode;
				moveItem.ItemDescription = item.ItemDescription;
				moveItem.Quantity = item.Quantity;
				moveItem.QuantityPerUom = item.QuantityPerUom;
				moveItem.Uom = item.Uom;
				moveItem.SourceLocationCode = item.LocationCode;
				moveItem.DestinationLocationCode = item.DestinationStorLocCode;
				moveItem.BBD = item.BestBeforeDate;
				moveItem.BatchNumber = item.BatchNumber;
				moveItem.ItemTransactionalInfoKey = item.ItemTransactionalInfoKey;

                result.Add(moveItem);
			}

			return result;
		}

		private static PmxOseWarehouseInfo ExecuteGetPmxWarehouseCode(PmxDbConnection conn, object[] parameters)
		{
			PmxOseProvider oseProv = new PmxOseProvider(conn);
			return oseProv.GetPmxWarehouseInfo((string)parameters[0]);
		}

		private static void DoDeliveryStockCreationBooking(WorkflowHelper methodHelper, BladeSwapDataCheckResult bladeSwapData, LocationResult locResult)
		{
			bool finished = false;
			int? luid = null;

			do
			{

				try
				{

					luid = methodHelper.SboProviderService.InvokeMethodWithDbConnection<int?>(true, false, null, null, delegate (PmxDbConnection conn, object[] parameters)
					{
						int? result = null;
						if (locResult.NeedSSCC)
						{
							PmxLogisticUnitIDProvider luidProv = new PmxLogisticUnitIDProvider(conn);
							string sscc;
							result = luidProv.GenerateNewLogisticUnitSSCC(null, out sscc);
						}
						else
						{
							result = locResult.LUID;
						}
						PmxDeliveryStockCreationQuorumProvider stockProv = new PmxDeliveryStockCreationQuorumProvider((SboConnection)conn);
						PmxDeliveryStockCreationQuorum deliveryStockBO = stockProv.GetNewBO();

						deliveryStockBO.LineStatus = Foundation.Data.Sbo.BusinessObjects.PmxDocumentLineStatus.Open;
						deliveryStockBO.LocationCode = locResult.DestinationLocationCode;
						deliveryStockBO.LUID = result;
						deliveryStockBO.OpenQuantity = bladeSwapData.BladeQuantityToSwap;
						deliveryStockBO.Quantity = bladeSwapData.BladeQuantityToSwap;
						deliveryStockBO.PickListDocEntry = bladeSwapData.PickListDocEntry;
						deliveryStockBO.SalesOrderDocEntry = bladeSwapData.SalesOrderDocEntry;
						deliveryStockBO.SalesOrderLineNumber = bladeSwapData.SalesOrderLineNum;

						stockProv.AddBO(deliveryStockBO);

						return result;
					});


					locResult.LUID = luid;

					finished = true;
				}
				catch (Exception ex)
				{
					s_log.Error("", ex);
					if (!(ex is ProdumexException) || ((ProdumexException)ex).Error == null)
					{
						// Inform the user with the error
						methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
							methodHelper.CultureInfo, BuildParamCollection(
								"MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, methodHelper.CultureInfo),
								"NoTranslationOfMessageKey", true,
								"ShowButton", true));
						methodHelper.WaitForMessage();

					}
					else
					{
						// Inform the user with the error
						methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
							methodHelper.CultureInfo, BuildParamCollection(
								"PmxError", ((ProdumexException)ex).Error,
								"ShowButton", true));
						methodHelper.WaitForMessage();
					}

					//retry?
					//ask user if he wants to retry
					methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
						methodHelper.CultureInfo, BuildParamCollection(
							"TitleKey", "MSG_TITLE_RETRY_UPDATE",
							"MessageKey", "MSG_QUESTION_AN_ERROR_OCCURED_WANT_RETRY"));
					methodHelper.Msg = methodHelper.WaitForMessage();

					if (methodHelper.Msg.Name.ToLower().EndsWith(".no"))
					{
						return;
					}
				}
			} while (!finished);
		}

		private static void PrintItemLabel(WorkflowHelper methodHelper, BladeSwapDataCheckResult bladeSwapData, LocationResult locResult)
		{
			bool next = true;
			do
			{
				PrintItemLabelFlowData printData = new PrintItemLabelFlowData()
				{
					BackRequested = false,
					ShowPrintMessage = true,
					LocationCode = locResult.DestinationLocationCode,
					NeededBatchNumber = null,
					NeededBatchNumber2 = null,
					NeededBestBeforeDate = null,
					NeededItemCode = bladeSwapData.BladeItemCodeToSwap,
					AskBatchInfo = false,
					EnteredQuantity = bladeSwapData.BladeQuantityToSwap
				};

				WorkflowInstanceBase printItemLabelFlow = methodHelper.Create("PrintItemLabelScript");

				printItemLabelFlow.BindParameters(
					methodHelper.CultureInfo,
					methodHelper.PmxOseCompany,
					new ReadOnlyBinder<string>("LocationCode", printData, "LocationCode"),
					new ReadOnlyBinder<string>("NeededItemCode", printData, "NeededItemCode"),
					new ReadOnlyBinder<string>("NeededBatchNumber", printData, "NeededBatchNumber"),
					new ReadOnlyBinder<string>("NeededBatchNumber2", printData, "NeededBatchNumber2"),
					new ReadOnlyBinder<DateTime?>("NeededBestBeforeDate", printData, "NeededBestBeforeDate"),
					new ReadOnlyBinder<int?>("NeededNrOfLabels", printData, "NeededNrOfLabels"),
					new ReadOnlyBinder<bool>("AskBatchInfo", printData, "AskBatchInfo"),
					new ReadWriteBinder<bool>("BackRequested", printData, "BackRequested"),
					new ReadWriteBinder<double?>("EnteredQuantity", printData, "EnteredQuantity"),
					new ReadWriteBinder<bool>("ShowPrintMessage", printData, "ShowPrintMessage")
					);

				methodHelper.StartSubFlowAndWait(printItemLabelFlow, true);

				next = !printData.BackRequested;

			} while (!next);
		}

		private static BladeSwapDataCheckResult CheckAndGatherDataForProcessing(WorkflowHelper methodHelper, Collection<PickingItemInfo> pickedItemInfos, int? luid, Collection<string> listOfSerialNumbers, int docEntry, string objType)
		{
			//First check if we need to perform actions

			//if (listOfSerialNumbers == null ||
			//	listOfSerialNumbers == null ||
			//	listOfSerialNumbers.Count == 0)
			//{
			//	//No serial numbers, so do nothing
			//	return new BladeSwapDataCheckResult(Steps.Stop);
			//}

			if (objType == null ||
			   objType == null ||
			   objType != Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.PmxObjectTypes.PmxWave)
			{
				//Not a wave, so do nothing
				return new BladeSwapDataCheckResult(Steps.Stop);
			}

			if (pickedItemInfos == null ||
			   pickedItemInfos == null ||
			   pickedItemInfos.Count == 0)
			{
				//No pick list object
				return new BladeSwapDataCheckResult(Steps.Stop);
			}

			int waveKey = docEntry;

			int pickListDocEntry = pickedItemInfos[0].PickListDocEntry;
			int pickListLineNum = pickedItemInfos[0].PickListLineID;

			string salesOrderQuery = BuildQuery.GetBladeInfoForPickList(waveKey, pickListDocEntry, pickListLineNum, methodHelper.SboProviderService.GetDbTool());
			DataSet salesOrderDs = methodHelper.SboProviderService.RunView(false, null, null, salesOrderQuery);

			if (GetNumberOfRows(salesOrderDs) == 0)
			{
				//Nothing for blade handling
				return new BladeSwapDataCheckResult(Steps.Stop);
			}

			string bladeSwapStatus = SystemStringConvertor.ObjectToStringOrNull(salesOrderDs.Tables[0].Rows[0][V33SalesOrderLinesTable.Columns.BladeSwap.NAME]);

			if (bladeSwapStatus != Definitions.BLADE)
			{
				return new BladeSwapDataCheckResult(Steps.Stop);
			}

			int salesOrderDocEntry = SystemInt32Convertor.ObjectToInt(salesOrderDs.Tables[0].Rows[0][V33SalesOrderLinesTable.Columns.DocEntry.NAME], 0, 0);
			int salesOrderLineNumber = SystemInt32Convertor.ObjectToInt(salesOrderDs.Tables[0].Rows[0][SboSalesOrderLineTable.Columns.LineNum.NAME], 0, 0);
            int salesOrderVisOrderNumber = SystemInt32Convertor.ObjectToInt(salesOrderDs.Tables[0].Rows[0][SboSalesOrderLineTable.Columns.VisOrder.NAME], 0, 0);
            double bladeInQuantityToSwap = SystemDoubleConvertor.ObjectToDouble(salesOrderDs.Tables[0].Rows[0][SboSalesOrderLineTable.Columns.Quantity.NAME], 0, 0);

            string fanQuery = BuildQuery.GetSalesOrderFanLine(salesOrderDocEntry, salesOrderLineNumber, salesOrderVisOrderNumber, methodHelper.SboProviderService.GetDbTool());
			DataSet fanDs = methodHelper.SboProviderService.RunView(false, null, null, fanQuery);

			if (GetNumberOfRows(fanDs) == 0)
			{
				return new BladeSwapDataCheckResult(Steps.Stop);
			}

			int fanVisOrderNum = SystemInt32Convertor.ObjectToInt(fanDs.Tables[0].Rows[0][SboSalesOrderLineTable.Columns.VisOrder.NAME], 0, 0);

			string bladeQuery = BuildQuery.GetSalesOrderSwapLine(salesOrderDocEntry, salesOrderLineNumber, fanVisOrderNum, methodHelper.SboProviderService.GetDbTool());
			DataSet bladeDs = methodHelper.SboProviderService.RunView(false, null, null, bladeQuery);

			if (GetNumberOfRows(bladeDs) == 0)
			{
				return new BladeSwapDataCheckResult(Steps.Stop);
			}

			double bladeOutQuantityToSwap = Math.Abs(SystemDoubleConvertor.ObjectToDouble(bladeDs.Tables[0].Rows[0][SboSalesOrderLineTable.Columns.Quantity.NAME], 0, 0));
            int bladeToSwapLineNumber = SystemInt32Convertor.ObjectToInt(bladeDs.Tables[0].Rows[0][SboSalesOrderLineTable.Columns.LineNum.NAME], 0, 0);
			string bladeItemCodeToSwap = SystemStringConvertor.ObjectToStringOrNull(bladeDs.Tables[0].Rows[0][SboSalesOrderLineTable.Columns.ItemCode.NAME]);
			string bladeItemDescriptionToSwap = SystemStringConvertor.ObjectToStringOrNull(bladeDs.Tables[0].Rows[0][PmxItemTable.Columns.CustomItemDescr.NAME]);
			string fanSwapInstrutions = SystemStringConvertor.ObjectToStringOrNull(fanDs.Tables[0].Rows[0][V33BladeSwapHeaderTable.Columns.Instructions.NAME]);
			string fanCode = SystemStringConvertor.ObjectToStringOrNull(fanDs.Tables[0].Rows[0][SboItemTable.Columns.ItemCode.NAME]);
			string fanDescription = SystemStringConvertor.ObjectToStringOrNull(fanDs.Tables[0].Rows[0][SboItemTable.Columns.ItemName.NAME]);

			return new BladeSwapDataCheckResult(Steps.Next, fanSwapInstrutions, fanCode, fanDescription, pickListDocEntry, salesOrderDocEntry, bladeToSwapLineNumber, bladeItemCodeToSwap,  bladeItemDescriptionToSwap, bladeInQuantityToSwap);
		}

		private static LocationResult GetDestinationLocation(WorkflowHelper methodHelper, List<MoveItemInfoExtended> moveItemInfoList, bool withLogisticCarrier,string  itemCode)
        {
            Step_IdentifyDestinationLocation:

            #region IDENTIFY DESTINATION LOCATION

            bool newSSCC = false;
            int? dstLuid = null;
            string targetLogCarrierCode = null;
            bool blockMoveWhenNotEmpty = false;
            bool useLocationSuggestions = false;
            string intermediateDestinationLocation = null;

            DataSet ds;

            Collection<LocationInfo> listOfLocationInfos = null;
            PmxDictionary<string, LocationInfo> listOfLocationInfosAsDictionary = null;
            BuildQuery.LocationFilter selectLocationFilter = BuildQuery.LocationFilter.Default;

            PmxItemInfo itemInfo = null;

            string whsCodeToUse = methodHelper.WhsInfo.PmxWarehouseCode;

            //Check if we have 1 distinct item code to move

            if (moveItemInfoList.Select(m => m.ItemCode).Distinct().Count() == 1)
            {
                itemInfo = methodHelper.SboProviderService.InvokeMethodWithDbConnection<PmxItemInfo>(false, false, null, null, new DelegateWithPmxDbConnection<PmxItemInfo>(ExecuteGetItemInfo), moveItemInfoList[0].ItemCode);
            }
            else
            {
                itemInfo = null;
            }

            try
            {
                //Get whether locations are suggested or not
                methodHelper.SboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
                    delegate (PmxDbConnection conn, object[] parameters)
                    {
                        PmxOseStorageLocationProvider slProv = new PmxOseStorageLocationProvider(conn);

                        LocationSuggestionRequest request = new LocationSuggestionRequest();

                        request.ListOfItemsToMove = moveItemInfoList.Cast<MoveItemInfo>().ToList();


                        listOfLocationInfos = slProv.GetSuggestedLocations(LocationRequestType.AdHocMove, request, whsCodeToUse, out useLocationSuggestions, out listOfLocationInfosAsDictionary);

                        return null;
                    });
            }
            catch (Exception ex)
            {
                //just log, and proceed as no suggestions were needed
                s_log.Error("Error getting location suggestions: ", ex);

                listOfLocationInfos = null;
            }

            	goto Step_IdentifyDestinationLocationAgain;

            #endregion

            Step_IdentifyDestinationLocationAgain:

            #region IDENTIFY DESTINATION LOCATION AGAIN

            string dstLocationCode = "";
            bool pressedSelectLocation = false;

            List<string> buttons = new List<string>();

            string defaultLocation = null ;
            itemInfo.DefaultWarehouseLocationOrZone.TryGetValue(whsCodeToUse, out defaultLocation);


            if (useLocationSuggestions)
            {
                buttons.Add("MSG_BUTTON_SELECT_OTHER_LOCATION");
                if (listOfLocationInfos == null ||
                    listOfLocationInfos.Count == 0)
                {
                    //No locations found, so show error
                    methodHelper.InitialErrorKey = "MSG_ERROR_NO_LOCATIONS_FOUND";
                    return new LocationResult(Steps.Return);
                }

                defaultLocation = listOfLocationInfos[0].LocationCode;
            }
            else
            {
                buttons.Add("MSG_BUTTON_SELECT_LOCATION");
                buttons.Add("MSG_BUTTON_SELECT_EMPTY_LOCATION");
            }

            if (!newSSCC)
            {
                buttons.Add("MSG_BUTTON_NEW_SSCC");
            }

            string scanDestinationLocationOrSSCC = string.Format(methodHelper.ResourceStore.GetValue<string>("MSG_TITLE_DST_LOCATION_OR_SSCC_FOR_ITEM"), itemCode);

            methodHelper.ShowScreen(typeof(IScanStringValueScreen),
                methodHelper.CultureInfo, BuildParamCollection(
                    "InitialErrorKey", methodHelper.InitialErrorKey,
                    "TitleKey", newSSCC ? "MSG_TITLE_SCAN_DESTINATION_LOCATION" : scanDestinationLocationOrSSCC,
                    "ForceDataEntry", true,
                    "AllowedButtonKeys", buttons.ToArray(),
                    "AllowToGoBack", true,
                    "Information", "MSG_LABEL_DEFAULT_LOCATION",
                    "MessageAddendum", defaultLocation

                    ));
            methodHelper.InitialErrorKey = null;
            methodHelper.Msg = methodHelper.WaitForMessage();

            if (methodHelper.Msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.IScanStringValueScreenReply.StringScanned")
            {
                string val = ExtractParameter<string>(methodHelper.Msg.Parameters, "stringValue");

                BarcodeContainer barcode = new BarcodeContainer();
                try
                {
                    barcode = methodHelper.BarcodeController.GetScannedData(val, true, barcode);
                }
                catch
                {
                    // Do nothing
                }
                if (!String.IsNullOrEmpty(barcode.GS1DataObject.Sscc))
                {
                    string query = BuildQuery.GetSSCCInfo(barcode.GS1DataObject.Sscc, null);
                    ds = methodHelper.SboProviderService.RunView(true, null, null, query);
                    if (GetNumberOfRows(ds) > 0)
                    {
                        newSSCC = false;

                        dstLuid = Convert.ToInt32(ds.Tables[0].Rows[0][SSCCDataSet.Columns.LUID.Name]);
                        dstLocationCode = Convert.ToString(ds.Tables[0].Rows[0][SSCCDataSet.Columns.STORAGE_LOCATION_CODE.Name]);
                    }
                    if (String.IsNullOrEmpty(dstLocationCode))
                    {
                        query = BuildQuery.GetMasterLuidOnSSCC(barcode.GS1DataObject.Sscc, methodHelper.SboProviderService.GetDbTool());
                        ds = methodHelper.SboProviderService.RunView(true, null, null, query);
                        if (GetNumberOfRows(ds) > 0)
                        {
                            methodHelper.InitialErrorKey = "MSG_ERROR_INCORRECT_SSCC_PARTIAL_TO_MASTER_SSCC";
                        }
                        else
                        {
                            methodHelper.InitialErrorKey = "MSG_ERROR_INCORRECT_SSCC";
                        }

                        goto Step_IdentifyDestinationLocation;
                    }
                    else if (!string.IsNullOrEmpty(intermediateDestinationLocation) &&
                        intermediateDestinationLocation != dstLocationCode)
                    {
                        methodHelper.InitialErrorKey = "MSG_ERROR_INCORRECT_LOCATION";
                        goto Step_IdentifyDestinationLocation;
                    }

                    goto Step_SelectDestinationLocation;
                }
                else
                {
                    dstLocationCode = val;

                    if (!string.IsNullOrEmpty(intermediateDestinationLocation) &&
                        intermediateDestinationLocation != dstLocationCode)
                    {
                        methodHelper.InitialErrorKey = "MSG_ERROR_INCORRECT_LOCATION";
                        goto Step_IdentifyDestinationLocation;
                    }

                    goto Step_SelectDestinationLocation;
                }
            }
            if (methodHelper.Msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.IScanStringValueScreenReply.ButtonClicked")
            {
                switch (ExtractParameter<string>(methodHelper.Msg.Parameters, "buttonKey"))
                {
                    case "MSG_BUTTON_SELECT_EMPTY_LOCATION":
                        selectLocationFilter = BuildQuery.LocationFilter.Empty;
                        break;
                    case "MSG_BUTTON_SELECT_NOT_FULL_LOCATION":
                        selectLocationFilter = BuildQuery.LocationFilter.NotFull;
                        break;
                    default:
                        selectLocationFilter = BuildQuery.LocationFilter.Default;
                        break;
                }

                string buttonKey = ExtractParameter<string>(methodHelper.Msg.Parameters, "buttonKey");
                if (buttonKey == "MSG_BUTTON_SELECT_LOCATION" || buttonKey == "MSG_BUTTON_SELECT_EMPTY_LOCATION" || buttonKey == "MSG_BUTTON_SELECT_OTHER_LOCATION")
                {
                    pressedSelectLocation = true;
                    goto Step_SelectDestinationLocation;
                }
                if (buttonKey == "MSG_BUTTON_NEW_SSCC")
                {
					if (withLogisticCarrier)
					{
						#region IDENTIFY LOGISTIC CARRIER

						string query = BuildQuery.GetLogisticCarriersOrderByName(true, true, methodHelper.SboProviderService.GetDbTool());
						DataSet dsLogisticCarriers = methodHelper.SboProviderService.RunView(false, null, null, query);
						if (GetNumberOfRows(dsLogisticCarriers) > 0)
						{
							methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreen),
								methodHelper.CultureInfo, BuildParamCollection(
								"InitialErrorKey", methodHelper.InitialErrorKey,
								"TitleKey", "MSG_TITLE_SELECT_LOGISTIC_CARRIER",
								"ProductDS", dsLogisticCarriers,
								"AllowNoSelection", false,
								"ButtonKey", "MSG_BUTTON_NO_LOGISTIC_CARRIER"));
							methodHelper.Msg = methodHelper.WaitForMessage();
							methodHelper.InitialErrorKey = null;

							if (methodHelper.Msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.ProductSelected")
							{
								targetLogCarrierCode = ExtractParameter<string>(methodHelper.Msg.Parameters, "itemCode");
							}
							else if (methodHelper.Msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.ButtonClicked")
							{
								targetLogCarrierCode = "";  // Remember we chose MSG_BUTTON_NO_LOGISTIC_CARRIER
							}
							else if (methodHelper.Msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.BackRequested")
							{
								goto Step_IdentifyDestinationLocation;
							}
							else if (methodHelper.Msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreenReply.StopRequested")
							{
								return new LocationResult(Steps.Stop);
							}
							else
							{
								methodHelper.InitialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
								goto Step_IdentifyDestinationLocation;
							}
						}

						#endregion
					}
					newSSCC = true;
                    goto Step_IdentifyDestinationLocationAgain;
                }
            }
            if (methodHelper.Msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.IScanStringValueScreenReply.BackRequested")
            {
                return new LocationResult(Steps.Return);
            }

            goto Step_IdentifyDestinationLocationAgain;

            #endregion

            Step_SelectDestinationLocation:

            #region SELECT DESTINATION LOCATION

            bool needsToEnterReasonForNotAllowedLocation = false;

            if (!useLocationSuggestions)
            {

                string query = BuildQuery.GetZoneTypesForItems(moveItemInfoList.Select(m => m.ItemCode).Distinct().ToList());
                ds = methodHelper.SboProviderService.RunView(true, null, null, query);
                List<string> zoneTypeCodes = null;
                if (GetNumberOfRows(ds) > 0)
                {
                    zoneTypeCodes = ds.Tables[0].AsEnumerable().Select(r => r.Field<string>(PmxItemZoneTypeTable.Columns.ZoneTypeCode.NAME)).Distinct().ToList();
                }

                query = BuildQuery.GetStorageLocationForZoneTypes(dstLocationCode, whsCodeToUse, zoneTypeCodes, selectLocationFilter, methodHelper.SboProviderService.GetDbTool());
                ds = methodHelper.SboProviderService.RunView(true, null, null, query);

            }
            else
            {
                //We are using suggested locations.
                //So now we need to check if the entered value is available in the list.

                if (!string.IsNullOrEmpty(dstLocationCode))
                {
                    //a location was entered, so check if it is available
                    if (listOfLocationInfosAsDictionary == null || !listOfLocationInfosAsDictionary.ContainsKey(dstLocationCode))
                    {
                        //Location not available
                        methodHelper.InitialErrorKey = "MSG_ERROR_LOCATION_NOT_AVAILABLE";
                        goto Step_IdentifyDestinationLocation;
                    }

                    //Location is available, so fill in the needed variables:
                    dstLocationCode = listOfLocationInfosAsDictionary[dstLocationCode].LocationCode;
                    blockMoveWhenNotEmpty = listOfLocationInfosAsDictionary[dstLocationCode].BlockMoveWhenNotEmpty;

                    //When the location is not an allowed suggested location, a reason needs to be entered
                    needsToEnterReasonForNotAllowedLocation = !listOfLocationInfosAsDictionary[dstLocationCode].AllowedSuggestedLocation;


                    goto Step_VerifyDestinationLocation;
                }
                else
                {
                    //no locations was entered, so build the list
                    ds = PmxOseStorageLocationProvider.ConvertLocationInfoListToLocationDataSet(listOfLocationInfos);
                }

            }

            if (GetNumberOfRows(ds) == 0)
            {
                methodHelper.InitialErrorKey = "MSG_ERROR_NO_LOCATION_FOUND";
                goto Step_IdentifyDestinationLocation;
            }
            if (!String.IsNullOrEmpty(dstLocationCode)
                && SystemInt32Convertor.ObjectToInt(ds.Tables[0].Rows[0][LocationDataSet.Columns.FREE_SPACE_FOR_LOG_UNITS.Name], 1, 1) < 1)
            {
                methodHelper.InitialErrorKey = "MSG_ERROR_LOCATION_NOT_AVAILABLE";
                goto Step_IdentifyDestinationLocation;
            }
            if (GetNumberOfRows(ds) == 1 &&
                !pressedSelectLocation)
            {
                //set locationcode from DB to avoid problems with uppercase/lowercase
                dstLocationCode = Convert.ToString(ds.Tables[0].Rows[0][LocationDataSet.Columns.CODE.Name]);
                blockMoveWhenNotEmpty = Convert.ToString(ds.Tables[0].Rows[0][PmxOseStorageLocationGeneralTable.Columns.BlockMoveWhenNotEmpty.NAME]) == SboBooleanColumnValues.Y.VALUE;

                goto Step_VerifyDestinationLocation;
            }

            methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreen),
                methodHelper.CultureInfo, BuildParamCollection(
                "InitialErrorKey", methodHelper.InitialErrorKey,
                "TitleKey", "MSG_TITLE_SELECT_DESTINATION_LOCATION",
                "LocationDS", ds));
            methodHelper.InitialErrorKey = null;
            methodHelper.Msg = methodHelper.WaitForMessage();

            if (methodHelper.Msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.LocationSelected")
            {
                dstLocationCode = ExtractParameter<string>(methodHelper.Msg.Parameters, "locationCode");
                DataRow[] rows = ds.Tables[0].Select(string.Format("{0} = '{1}'", LocationDataSet.Columns.CODE.Name, dstLocationCode));

                if (rows.Length > 0)
                {
                    blockMoveWhenNotEmpty = Convert.ToString(rows[0][PmxOseStorageLocationGeneralTable.Columns.BlockMoveWhenNotEmpty.NAME]) == SboBooleanColumnValues.Y.VALUE;
                }
                goto Step_VerifyDestinationLocation;
            }

            goto Step_IdentifyDestinationLocation;

            #endregion

            Step_VerifyDestinationLocation:

            #region VERIFY DESTINATION LOCATION

            if (useLocationSuggestions)
            {
                //For suggested locations, we need to check if we need to enter a reason
                //A location was entered, so check if it is available
                if (listOfLocationInfosAsDictionary == null || !listOfLocationInfosAsDictionary.ContainsKey(dstLocationCode))
                {
                    //Location not available
                    methodHelper.InitialErrorKey = "MSG_ERROR_LOCATION_NOT_AVAILABLE";
                    goto Step_IdentifyDestinationLocation;
                }

                //When the location is not an allowed suggested location, a reason needs to be entered
                needsToEnterReasonForNotAllowedLocation = !listOfLocationInfosAsDictionary[dstLocationCode].AllowedSuggestedLocation;

            }


            if (!string.IsNullOrEmpty(intermediateDestinationLocation) &&
                        intermediateDestinationLocation != dstLocationCode)
            {
                methodHelper.InitialErrorKey = "MSG_ERROR_INCORRECT_LOCATION";
                goto Step_IdentifyDestinationLocation;
            }



            bool isLocked = methodHelper.SboProviderService.InvokeMethodWithDbConnection<bool>(false, true, null, null, delegate (PmxDbConnection conn, object[] parameters)
            {
                PmxOseStorageLocationProvider prov = new PmxOseStorageLocationProvider(conn);
                return prov.IsLocationLocked(dstLocationCode);
            });

            if (isLocked)
            {
                methodHelper.InitialErrorKey = "MSG_ERROR_DESTINATION_LOCATION_LOCKED";
                goto Step_IdentifyDestinationLocation;
            }

            if (blockMoveWhenNotEmpty)
            {
                bool locationEmpty = methodHelper.SboProviderService.InvokeMethodWithDbConnection<bool>(false, true, null, null, delegate (PmxDbConnection conn, object[] parameters)
                {
                    PmxInventoryProvider prov = new PmxInventoryProvider(conn);
                    double? dummy;
                    double tempQty = prov.GetCurrentQuantityForLocation(dstLocationCode, out dummy);

                    return tempQty == 0;
                });

                if (!locationEmpty)
                {
                    methodHelper.InitialErrorKey = "MSG_ERROR_ITEMS_CANNOT_BE_STORED_ON_LOCATION_NOT_EMPTY";
                    goto Step_IdentifyDestinationLocation;
                }
            }


            PmxDictionary<string, PmxStorageLocationCanItemBeStoredInfo> canItemBeStoredInfoPerItemCode = new PmxDictionary<string, PmxStorageLocationCanItemBeStoredInfo>();

            foreach (MoveItemInfoExtended info in moveItemInfoList)
            {
                PmxStorageLocationCanItemBeStoredInfo entry;
                if (canItemBeStoredInfoPerItemCode.TryGetValue(info.ItemCode, out entry))
                {
                    entry.Quantity += info.Quantity;
                }
                else
                {
                    entry = new PmxStorageLocationCanItemBeStoredInfo(info.Quantity, null, null);
                    canItemBeStoredInfoPerItemCode.Add(info.ItemCode, entry);
                }
                if (info.ItemTransactionalInfoKey.HasValue)
                {
                    if (entry.ItriKeys == null)
                        entry.ItriKeys = new List<int>();
                    entry.ItriKeys.Add(info.ItemTransactionalInfoKey.Value);
                }
            }


            bool canBeStored = methodHelper.SboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null, delegate (PmxDbConnection conn, object[] parameters)
            {
                PmxOseStorageLocationProvider prov = new PmxOseStorageLocationProvider(conn);
                return prov.CanItemsBeStoredOnStorageLocation(dstLocationCode, canItemBeStoredInfoPerItemCode);
            });
            if (!canBeStored)
            {
                //items cannot be stored here
                //show message
                methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    methodHelper.CultureInfo,
                    BuildParamCollection(
                        "MessageKey", "MSG_ERROR_STORAGE_LOCATION_NOT_ALLOWED_FOR_ITEM",
                        "ShowButton", true));
                methodHelper.Msg = methodHelper.WaitForMessage();


                goto Step_IdentifyDestinationLocation;

            }

            foreach (KeyValuePair<string, PmxStorageLocationCanItemBeStoredInfo> kv in canItemBeStoredInfoPerItemCode)
            {
                if (kv.Value.Status == PmxStorageLocationCanItemBeStored.BlockDifferentItemOrBatch)
                {
                    // Inform the user with the error
                    methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        methodHelper.CultureInfo, BuildParamCollection(
                            "MessageKey", "MSG_BLOCK_ADDING_DIFFERENT_ITEM_OR_BATCH",
                            "NoTranslationOfMessageKey", false,
                            "IsError", true,
                            "ShowButton", true));
                    methodHelper.WaitForMessage();

                    goto Step_IdentifyDestinationLocation;
                }
            }

            foreach (KeyValuePair<string, PmxStorageLocationCanItemBeStoredInfo> kv in canItemBeStoredInfoPerItemCode)
            {
                if (kv.Value.Status != PmxStorageLocationCanItemBeStored.Yes)
                {
                    // Warn user and ask to continue
                    methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                        methodHelper.CultureInfo, BuildParamCollection(
                            "TitleKey", "MSG_TITLE_CONFIRM_MOVE",
                            "MessageKey", "MSG_CONFIRM_ADDING_DIFFERENT_ITEM_OR_BATCH"));
                    methodHelper.Msg = methodHelper.WaitForMessage();

                    if (!methodHelper.Msg.Name.EndsWith("IDecisionScreenReply.Yes"))
                    {
                        goto Step_IdentifyDestinationLocation;
                    }
                    break;
                }
            }

            //Verify if there are serial numbers that require an LUID
            if (!newSSCC && !dstLuid.HasValue)
            {

                List<string> itemCodes = moveItemInfoList.Select(m => m.ItemCode).Distinct().ToList();
                PmxDictionary<string, PmxItemInfo> listOfItemInfos = null;

                methodHelper.SboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null, delegate (PmxDbConnection conn, object[] parameters)
                {
                    PmxItemAllConnectionsProvider prov = new PmxItemAllConnectionsProvider(conn);
                    listOfItemInfos = prov.GetCachedItemInfoInDictionary(itemCodes);
                    return null;
                });

                foreach (MoveItemInfoExtended info in moveItemInfoList)
                {
                    //check if we have new SSCC when moving serial number with location tracking
                    if (listOfItemInfos[info.ItemCode].TrackLocationSerialNumber)
                    {
                        if (!info.DestinationLogisticUnitIdentKey.HasValue)
                        {
                            methodHelper.InitialErrorKey = "MSG_ERROR_SERIAL_NUMMER_WITH_TRACKING_LOCATION_NEEDS_LUID";
                            goto Step_IdentifyDestinationLocation;
                        }

                    }
                }

            }
            goto Step_EnterReason;

            #endregion

            Step_EnterReason:

            #region ENTER REASON

            string reasonCode = null;
            string reasonFreeText = null;
            string reasonLocationCode = null;
            string reasonName = null;

            if (needsToEnterReasonForNotAllowedLocation)
            {

                if (listOfLocationInfos != null &&
                    listOfLocationInfos.Count > 0)
                {
                    reasonLocationCode = listOfLocationInfos[0].LocationCode;
                }

                //get all reasons for deviating standard flows
                string query = BuildQuery.GetReasonsForDeviatingFromStandardProcess();
                DataSet dsReasons = methodHelper.SboProviderService.RunView(false, null, null, query);



                methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectReasonScreen),
                          methodHelper.CultureInfo, BuildParamCollection(
                          "InitialErrorKey", methodHelper.InitialErrorKey,
                          "TitleKey", "MSG_TITLE_SELECT_REASON",
                          "MessageKey", "MSG_SELECT_REASON_FOR_SELECTING_OTHER_LOCATION",
                          "ReasonDS", dsReasons));
                methodHelper.Msg = methodHelper.WaitForMessage();
                methodHelper.InitialErrorKey = null;

                if (methodHelper.Msg.Name != "Produmex.Foundation.SlimScreen.Interfaces.ISelectReasonScreenReply.ReasonSelected")
                {
                    //go back
                    goto Step_IdentifyDestinationLocation;
                }

                reasonCode = ExtractParameter<string>(methodHelper.Msg.Parameters, "reasonCode");

                bool enterFreeTextReason = false;

                DataRow[] rows = dsReasons.Tables[0].Select(String.Format("{0} = '{1}'", ReasonDataSet.Columns.REASON_CODE.Name, reasonCode));
                if (rows.Length > 0)
                {
                    reasonName = rows[0][ReasonDataSet.Columns.REASON_NAME.Name] as string;
                    enterFreeTextReason = rows[0][ReasonDataSet.Columns.REASON_REQUIRES_FREE_TEXT.Name] as string == SboBooleanColumnValues.Y.VALUE;
                }

                // Show screen with list
                #region ENTER REASON FREE TEXT

                if (enterFreeTextReason)
                {

                    // Show screen with list
                    methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IEnterStringValueScreen),
                        methodHelper.CultureInfo, BuildParamCollection(
                            "InitialErrorKey", methodHelper.InitialErrorKey,
                            "TitleKey", "MSG_TITLE_ENTER_REASON",
                            "Information", "MSG_ENTER_FREE_REASON_DEFAULT",
                            "Parameters", new object[] { reasonName },
                            "AllowToGoBack", true,
                            "ForceDataEntry", true,
                            "AllowMultiLine", true,
                            "MinimumNumberOfCharacters", Definitions.MIMIMUM_NUMBER_OF_CHARACTERS_FOR_REASON
                            ));

                    methodHelper.Msg = methodHelper.WaitForMessage();

                    methodHelper.InitialErrorKey = null;

                    if (methodHelper.Msg.Name.EndsWith(".StringEntered"))
                    {
                        reasonFreeText = ExtractParameter<string>(methodHelper.Msg.Parameters, "stringValue");

                    }
                    else
                    {
                        //go back
                        goto Step_EnterReason;
                    }


                }
                #endregion
            }

            return new LocationResult(Steps.Next, dstLocationCode, dstLuid, newSSCC, targetLogCarrierCode);

            #endregion
        }

        private static void PrintSSCCLabel(WorkflowHelper methodHelper, bool newSSCC, string dstLocationCode, string deviceId, int? dstLuid)
        { 
            if (newSSCC)
            {
                int numberOfReports = 0;
                bool askReprint = false;
                bool proceed = false;

                while (!proceed)
                {
                    numberOfReports = 0;

                    try
                    {
                        //trigger event
                        numberOfReports += methodHelper.SboProviderService.InvokeMethodWithDbConnection<int>(false, true, null, null, delegate (PmxDbConnection conn, object[] parameters)
                        {
                            PmxPrintReportProvider printProv = new PmxPrintReportProvider(conn);
                            return printProv.PrintReportsForEvent(PmxPrintReportEventType.WarehouseCreatedLogisticUnit, dstLocationCode, deviceId, dstLuid.Value, out askReprint);
                        });
                    }
                    catch (Produmex.Foundation.ProdumexException ex)
                    {
                        s_log.Error(ex.ToString());
                        methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            methodHelper.CultureInfo, BuildParamCollection(
                                "InitialErrorKey", methodHelper.InitialErrorKey,
                                "MessageKey", KnownMessageKeys.MSG_ERROR_INTERNAL_ERROR,
                                "MessageAddendum", ex.Message,
                                "PmxError", ex.Error,
                                "ShowButton", true));

                        methodHelper.Msg = methodHelper.WaitForMessage();
                    }
                    catch (Exception ex)
                    {
                        s_log.Error(ex.ToString());
                        methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                            methodHelper.CultureInfo, BuildParamCollection(
                                "InitialErrorKey", methodHelper.InitialErrorKey,
                                "MessageKey", KnownMessageKeys.MSG_ERROR_INTERNAL_ERROR,
                                "MessageAddendum", ex.Message,
                                "PmxError", Produmex.Sbo.Logex.Data.ProdumexErrors.UnexpectedError(s_log, ex),
                                "ShowButton", true));

                        methodHelper.Msg = methodHelper.WaitForMessage();
                    }

                    if (numberOfReports > 0 && askReprint)
                    {
                        methodHelper.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                            methodHelper.CultureInfo, BuildParamCollection(
                                "MessageKey", "MSG_QUESTION_REPRINT_LABEL",
                                "TitleKey", "MSG_TITLE_REPRINT_LABEL"
                                ));

                        methodHelper.Msg = methodHelper.WaitForMessage();

                        if (methodHelper.Msg.Name.EndsWith(".No"))
                        {
                            proceed = true;
                        }

                    }
                    else
                    {
                        proceed = true;
                    }

                }
            }
        }


        #endregion

        #region HELPERS

        private enum Steps
        {
            Return,
            Next,
			Stop
		}

		private class PrintItemLabelFlowData
		{
			public bool BackRequested;
			public string LocationCode;
			public string NeededItemCode;
			public string NeededBatchNumber;
			public string NeededBatchNumber2;
			public DateTime? NeededBestBeforeDate;
			public int? NeededNrOfLabels;
			public bool AskBatchInfo;
			public double? EnteredQuantity;
			public bool ShowPrintMessage;
		}

		private class BladeSwapDataCheckResult
		{
			public BladeSwapDataCheckResult(Steps step) : this(step, null, null, null, 0, 0, 0, null, null, 0)
			{
			}
			public BladeSwapDataCheckResult(Steps step, string fanSwapInstructions, string fanCode, string fanDescription, int picklistDocEntry, int salesOrderDocEntry, int salesOrderLineNum, string bladeItemCodeToSwap, string bladeItemDescriptionToSwap, double bladeQuantityToSwap)
			{
				this.Step = step;
				this.FanSwapInstructions = fanSwapInstructions;
				this.FanCode = fanCode;
				this.FanDescription = fanDescription;
				this.PickListDocEntry = picklistDocEntry;
				this.SalesOrderDocEntry = salesOrderDocEntry;
				this.BladeQuantityToSwap = bladeQuantityToSwap;
				this.BladeItemCodeToSwap = bladeItemCodeToSwap;
				this.BladeItemDescriptionToSwap = bladeItemDescriptionToSwap;
				this.SalesOrderLineNum = salesOrderLineNum;
			}

			public double BladeQuantityToSwap;
			public string FanCode;
			public string FanDescription;
			public string FanSwapInstructions;
			public int PickListDocEntry;
			public int SalesOrderDocEntry;
			public Steps Step;
			public string BladeItemCodeToSwap;
			public int SalesOrderLineNum;
			public string BladeItemDescriptionToSwap;
		}

		private class LocationResult
		{
			public LocationResult(Steps step) : this(step, null, null, false, null)
			{
			}
			public LocationResult(Steps step, string locationCode, int? luid, bool needSSCC, string logisticCarrier)
			{
				this.Step = step;
				this.DestinationLocationCode = locationCode;
				this.LUID = luid;
				this.LogisticCarrier = logisticCarrier;
                this.NeedSSCC = needSSCC;
			}

			public string DestinationLocationCode;
			public string LogisticCarrier;
			public int? LUID;
			public Steps Step;
            public bool NeedSSCC;
		}

        private class WorkflowHelper
        {
            public ISboProviderService SboProviderService;
            public Action<Type, CultureInfo, ReadOnlyCollection<Parameter>> ShowScreen;
            public Func<Message> WaitForMessage;
			public Func<string, WorkflowInstanceBase> Create;
			public Action<WorkflowInstanceBase, bool> StartSubFlowAndWait;
			public ReadOnlyBinder<CultureInfo> CultureInfo;
            public Message Msg;
            public string InitialErrorKey;
            internal IBarcodeController BarcodeController;
			public string DeviceInstanceId;
			public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;
			public PmxOseWarehouseInfo WhsInfo;        
            public IResourceStore ResourceStore;
        }

        private static int GetNumberOfRows(DataSet ds)
        {
            if (ds != null && ds.Tables.Count > 0)
            {
                return ds.Tables[0].Rows.Count;
            }
            return 0;
        }

        private static PmxItemInfo ExecuteGetItemInfo(PmxDbConnection conn, object[] parameters)
        {
            PmxItemAllConnectionsProvider prov = new PmxItemAllConnectionsProvider(conn);
            return prov.GetCachedItemInfo((string)parameters[0]);
        }
        #endregion

        #region QUERY

        /// <summary>
        /// Static class grouping our native queries
        /// </summary>
        private class BuildQuery
        {
            internal enum LocationFilter
            {
                All = 0,
                Empty = 1,
                NotFull = 2,
                NotFullAndIgnoreNonLuidStock = 3,

                Default = NotFull
            }

            internal static string GetSSCCInfo(string sscc, List<int> listOfLUID)
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.AppendFormat("SELECT DISTINCT {0}",
                    SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME, SSCCDataSet.Columns.LUID.Name));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.SSCC.NAME, SSCCDataSet.Columns.SSCC.Name));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME, SSCCDataSet.Columns.STORAGE_LOCATION_CODE.Name));
                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME, SSCCDataSet.Columns.PMX_WAREHOUSE_CODE.Name));

                strBuilder.AppendFormat(", {0}",
                    SqlCommandHelper.EscapeTableColumnNameWithAlias(PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Name.NAME, SSCCDataSet.Columns.PMX_WAREHOUSE_NAME.Name));


                strBuilder.AppendFormat("\r\n FROM {0}",
                    SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));

                strBuilder.AppendFormat("\r\n INNER JOIN {0}",
                    SqlCommandHelper.EscapeTableName(PmxLogisticUnitIDTable.NAME));

                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME));

                strBuilder.AppendFormat("\r\n INNER JOIN {0}",
                SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));

                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME));

                strBuilder.AppendFormat("\r\n INNER JOIN {0}",
                    SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.AppendFormat(" ON {0} = {1}",
                    SqlCommandHelper.EscapeTableColumnName(PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME));



                string operand = "\r\n WHERE ";

                if (!string.IsNullOrEmpty(sscc))
                {
                    strBuilder.Append(operand);
                    strBuilder.AppendFormat(" {0} = {1}",
                        SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.SSCC.NAME),
                        SqlCommandHelper.FormatAndEscapeColumnValue(sscc, CultureInfo.InvariantCulture));
                    operand = "\r\n AND ";
                }


                if (listOfLUID != null && listOfLUID.Count > 0)
                {
                    strBuilder.Append(operand);
                    strBuilder.AppendFormat(" {0} IN ({1})",
                        SqlCommandHelper.EscapeTableColumnName(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME),
                        SqlCommandHelper.FormatAndEscapeColumnValue(listOfLUID, CultureInfo.InvariantCulture));
                    operand = "\r\n AND ";
                }

                return strBuilder.ToString();
            }

            /// <summary>
            /// Gets the master luid configuration SSCC.
            /// </summary>
            /// <param name="sscc">The SSCC.</param>
            /// <param name="dbTool">The database tool.</param>
            /// <returns></returns>
            public static string GetMasterLuidOnSSCC(string sscc, DbTool dbTool)
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();

                strBuilder.Append("SELECT DISTINCT ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxLinkMasterLUIDTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxLinkMasterLUIDTable.Columns.MasterLUID.NAME));

                strBuilder.Append("\r\n FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxLinkMasterLUIDTable.NAME));

                strBuilder.Append("\r\n INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxLogisticUnitIDTable.NAME));
                strBuilder.Append(" ON ", strBuilder.AsIdentifier(PmxLogisticUnitIDTable.NAME, PmxLogisticUnitIDTable.Columns.InternalKey.NAME));
                strBuilder.Append(" = ", strBuilder.AsIdentifier(PmxLinkMasterLUIDTable.NAME, PmxLinkMasterLUIDTable.Columns.MasterLUID.NAME));

                strBuilder.Append("\r\n WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxLogisticUnitIDTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxLogisticUnitIDTable.Columns.SSCC.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(sscc, CultureInfo.InvariantCulture));


                return strBuilder.ToString();
            }

            /// <summary>
            /// Gets the name and the code of the  location, ordered by name.
            /// </summary>
            /// <param name="locationCode">The location code.</param>
            /// <param name="warehouseCode">The warehouse code.</param>
            /// <param name="zoneTypeCodes">The zone type codes.</param>
            /// <param name="locationFilter">The location filter.</param>
            /// <param name="dbTool">The db tool.</param>
            /// <returns>The sql statement to get the location</returns>
            internal static string GetStorageLocationForZoneTypes(string locationCode, string warehouseCode, IList<string> zoneTypeCodes, LocationFilter locationFilter, DbTool dbTool)
            {
                bool isLookup = !String.IsNullOrEmpty(locationCode);
                if (isLookup)
                {
                    // This is a lookup, extra info will be returned instead of filtering
                    locationFilter = LocationFilter.All;
                }

                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();
                strBuilder.Append("SELECT ");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME, LocationDataSet.Columns.CODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Name.NAME, LocationDataSet.Columns.DESCRIPTION.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME, LocationDataSet.Columns.WAREHOUSECODE.Name);

                //strBuilder.Append(", ");
                //strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                //strBuilder.Append(".");
                //strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.BlockMoveWhenNotEmpty.NAME));

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Name.NAME, LocationDataSet.Columns.WAREHOUSENAME.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.BlockMoveWhenNotEmpty.NAME));

                if (isLookup)
                {
                    // locationCode is provided, this is a lookup, return some info about the location

                    strBuilder.AppendFormat(", CASE COUNT( {0} ) WHEN 0 THEN 1 ELSE 0 END AS {1}",
                        SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.InternalKey.NAME),
                        SqlCommandHelper.EscapeColumnName(LocationDataSet.Columns.IS_EMPTY.Name));
                    strBuilder.AppendFormat(", {0} - COUNT( DISTINCT {1}) AS {2}",
                        SqlCommandHelper.EscapeTableColumnName(PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.MaximumLogisticUnits.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME),
                        SqlCommandHelper.EscapeColumnName(LocationDataSet.Columns.FREE_SPACE_FOR_LOG_UNITS.Name));
                }


                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);


                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseWarehouseTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Code.NAME);

                strBuilder.Append("\r\n INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxWarehouseTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxWarehouseTable.NAME, PmxWarehouseTable.Columns.WarehouseCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.SboWarehouseCode.NAME);

                //strBuilder.Append("\r\n AND ");
                //SqlCommandHelper.AddTableColumnName(strBuilder, PmxWarehouseTable.NAME, PmxWarehouseTable.Columns.IsManagedByProdumex.NAME);
                //strBuilder.Append(" = ");
                //strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));



                if (isLookup || locationFilter != LocationFilter.All)
                {
                    strBuilder.Append(" LEFT JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(" ON ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME);
                    strBuilder.Append(" = ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

                    if (locationFilter == LocationFilter.NotFullAndIgnoreNonLuidStock)
                    {
                        strBuilder.AppendFormat(" AND {0} IS NOT NULL",
                            SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME));
                    }
                }

                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.IsActive.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));


                if (isLookup)
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Code.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(locationCode, CultureInfo.InvariantCulture));
                }

                if (warehouseCode != null &&
                    warehouseCode.Trim() != "")
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(warehouseCode, CultureInfo.InvariantCulture));

                }

                if (zoneTypeCodes != null &&
                    zoneTypeCodes.Count > 0)
                {
                    foreach (string zoneTypeCode in zoneTypeCodes)
                    {
                        strBuilder.Append(" AND ");
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                        strBuilder.Append(".");
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Code.NAME));
                        strBuilder.Append(" IN ( ");

                        strBuilder.Append("SELECT ");

                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                        strBuilder.Append(".");
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Code.NAME));


                        strBuilder.Append(" FROM ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));
                        strBuilder.Append(" INNER JOIN ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                        strBuilder.Append(" ON ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                        strBuilder.Append(" = ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);
                        strBuilder.Append(" RIGHT JOIN ");
                        strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxLinkZoneZoneTypeTable.NAME));
                        strBuilder.Append(" ON ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxZoneCode.NAME);
                        strBuilder.Append(" = ");
                        SqlCommandHelper.AddTableColumnName(strBuilder, PmxLinkZoneZoneTypeTable.NAME, PmxLinkZoneZoneTypeTable.Columns.ParentCode.NAME);


                        strBuilder.Append(" WHERE ");


                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxLinkZoneZoneTypeTable.NAME));
                        strBuilder.Append(".");
                        strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxLinkZoneZoneTypeTable.Columns.ZoneTypeCode.NAME));
                        strBuilder.Append(" = ");
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(zoneTypeCode, CultureInfo.InvariantCulture));


                        strBuilder.Append(") ");

                    }
                }

                if (locationFilter == LocationFilter.Empty)
                {
                    //If there is no link to the inventory, the location is empty
                    strBuilder.Append(" AND ");
                    SqlCommandHelper.AddTableColumnName(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME);
                    strBuilder.Append(" IS NULL ");
                }

                if (isLookup
                    || locationFilter == LocationFilter.NotFull
                    || locationFilter == LocationFilter.NotFullAndIgnoreNonLuidStock)
                {
                    strBuilder.AppendFormat("\r\n GROUP BY {0}",
                        SqlCommandHelper.EscapeTableColumnName(PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME));
                    strBuilder.AppendFormat(", {0}",
                        SqlCommandHelper.EscapeTableColumnName(PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Name.NAME));
                    strBuilder.AppendFormat(", {0}",
                        SqlCommandHelper.EscapeTableColumnName(PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME));
                    strBuilder.AppendFormat(", {0}",
                        SqlCommandHelper.EscapeTableColumnName(PmxOseWarehouseTable.NAME, PmxOseWarehouseTable.Columns.Name.NAME));
                    strBuilder.AppendFormat(", {0}",
                       SqlCommandHelper.EscapeTableColumnName(PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.BlockMoveWhenNotEmpty.NAME));
                    strBuilder.AppendFormat(", {0}",
                        SqlCommandHelper.EscapeTableColumnName(PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.BlockMoveWhenNotEmpty.NAME));
                    strBuilder.AppendFormat(", {0}",
                        SqlCommandHelper.EscapeTableColumnName(PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.MaximumLogisticUnits.NAME));
                }

                if (locationFilter == LocationFilter.NotFull)
                {
                    strBuilder.AppendFormat("\r\n HAVING ( COUNT( DISTINCT {0} ) < {1} OR {1} IS NULL )",
                        SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME),
                        SqlCommandHelper.EscapeTableColumnName(PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.MaximumLogisticUnits.NAME));
                }

                if (!isLookup)
                {
                    strBuilder.AppendFormat("\r\n ORDER BY {0}",
                        SqlCommandHelper.EscapeTableColumnName(PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Name.NAME));
                }

                return strBuilder.ToString();
            }


            /// <summary>
            /// Gets the zone types for items.
            /// </summary>
            /// <param name="itemCodes">The item codes.</param>
            /// <returns></returns>
            public static string GetZoneTypesForItems(List<string> itemCodes)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT DISTINCT");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxItemZoneTypeTable.NAME, PmxItemZoneTypeTable.Columns.ZoneTypeCode.NAME, PmxItemZoneTypeTable.Columns.ZoneTypeCode.NAME);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemZoneTypeTable.NAME));


                string operand = "";
                strBuilder.Append(" WHERE ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemZoneTypeTable.NAME, PmxItemZoneTypeTable.Columns.ItemCode.NAME);
                strBuilder.Append(" IN ( ");

                foreach (string itemCode in itemCodes)
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));
                    operand = ", ";
                }
                strBuilder.Append(") ");


                return strBuilder.ToString();
            }

            /// <summary>
            /// Gets the reasons.
            /// </summary>
            /// <returns></returns>
            public static string GetReasonsForDeviatingFromStandardProcess()
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.Append("SELECT ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxReasonTable.NAME, PmxReasonTable.Columns.Code.NAME, ReasonDataSet.Columns.REASON_CODE.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxReasonTable.NAME, PmxReasonTable.Columns.Name.NAME, ReasonDataSet.Columns.REASON_NAME.Name);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxReasonTable.NAME, PmxReasonTable.Columns.RequiresFreeText.NAME, ReasonDataSet.Columns.REASON_REQUIRES_FREE_TEXT.Name);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxReasonTable.NAME));

                strBuilder.Append(" WHERE ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxReasonTable.NAME, PmxReasonTable.Columns.UseWhenDeviatingFromStandardProcess.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.CurrentCulture));

                strBuilder.AppendFormat(
                    " ORDER BY CASE WHEN {0} IS NULL THEN 1 ELSE 0 END, {0} ",
                     SqlCommandHelper.EscapeTableColumnName(PmxReasonTable.NAME, PmxReasonTable.Columns.SequenceNumber.NAME));

                return strBuilder.ToString();
            }

            /// <summary>
            /// Command for getting all logistic carriers order by name
            /// </summary>
            /// <param name="onlyInStock">if set to <c>true</c> the system will get the logistic carriers that are in stock and have no products on it (LUID IS NULL)</param>
            /// <param name="filterOutFrozenItems">if set to <c>true</c> [filter out frozen items].</param>
            /// <returns>
            /// Command for getting all logistic carriers order by name
            /// </returns>
            public static string GetLogisticCarriersOrderByName(bool onlyInStock, bool filterOutFrozenItems, DbTool dbTool)
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();

                strBuilder.Append("SELECT DISTINCT ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME, ProductDataSet.Columns.PRODUCT_CODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, SboItemTable.NAME, PmxItemTable.Columns.CustomItemDescr.NAME, ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, SboItemTable.NAME, SboItemTable.Columns.CodeBars.NAME, ProductDataSet.Columns.GTIN.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.IsLogisticCarrier.NAME));

                strBuilder.Append("\r\n FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));

                if (onlyInStock)
                {
                    strBuilder.Append("\r\n INNER JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                    strBuilder.Append(" ON ", strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" = ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" AND ", strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME));
                    strBuilder.Append(" IS NULL ");
                }

                strBuilder.Append("\r\n WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.InventoryItem.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.IsLogisticCarrier.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                if (filterOutFrozenItems)
                {
                    Produmex.Sbo.Logex.Data.SboCommands.GetFrozenItemFilterSQLWherePart(strBuilder, " AND ");
                }

                strBuilder.Append("\r\n ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.CustomItemDescr.NAME));

                return strBuilder.ToString();
            }

            public static string GetBladeInfoForPickList(int wavekey, int pickListDocEntry, int pickListLineNum, DbTool dbtool)
            {
                DbQueryBuilder strBuilder = dbtool.GetQueryBuilder();

                strBuilder.Append("SELECT ");

                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.DocEntry.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.LineNum.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.VisOrder.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, V33SalesOrderLinesTable.NAME, V33SalesOrderLinesTable.Columns.BladeSwap.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.Quantity.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListTable.NAME));


                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListLineTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxPickListProposalLineTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.DocEntry.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.BaseEntry.NAME);

                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.BaseLine.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.LineNum.NAME);

                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboSalesOrderLineTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.DocEntry.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseEntry.NAME);
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.LineNum.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseLine.NAME);
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.ObjectType.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseType.NAME);


                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListTable.Columns.WaveKey.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(wavekey, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListTable.Columns.DocEntry.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(pickListDocEntry, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListLineTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxPickListLineTable.Columns.LineNum.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(pickListLineNum, CultureInfo.InvariantCulture));

                return strBuilder.ToString();
            }

            public static string GetSalesOrderFanLine(int salesOrderDocEntry, int salesOrderLineNumber, int salesOrderVisOrderNumber, DbTool dbTool)
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();
                strBuilder.Append("SELECT TOP 1");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.DocEntry.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.ItemCode.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemName.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.Quantity.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.VisOrder.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, V33SalesOrderLinesTable.NAME, V33SalesOrderLinesTable.Columns.BladeSwap.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, V33BladeSwapHeaderTable.NAME, V33BladeSwapHeaderTable.Columns.Instructions.NAME);

                strBuilder.Append("\n FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboSalesOrderLineTable.NAME));

                strBuilder.Append("\n LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(V33BladeSwapHeaderTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.ItemCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, V33BladeSwapHeaderTable.NAME, V33BladeSwapHeaderTable.Columns.Code.NAME);

                strBuilder.Append("\n LEFT JOIN ");
                strBuilder.Append(SboItemTable.NAME);
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.ItemCode.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemName.NAME);

                strBuilder.Append("\n WHERE ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.DocEntry.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(salesOrderDocEntry, CultureInfo.InvariantCulture));
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.VisOrder.NAME);
                strBuilder.Append(" < ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(salesOrderVisOrderNumber, CultureInfo.InvariantCulture));
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, V33SalesOrderLinesTable.NAME, V33SalesOrderLinesTable.Columns.BladeSwap.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue("Y", CultureInfo.InvariantCulture));
                strBuilder.Append(" ORDER BY ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.VisOrder.NAME);
                strBuilder.Append(" DESC ");

                return strBuilder.ToString();
            }

            public static string GetSalesOrderSwapLine(int salesOrderDocEntry, int salesOrderLineNumber, int salesOrderVisOrderNumber, DbTool dbTool)
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();
                strBuilder.Append("SELECT TOP 1");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.DocEntry.NAME);
				strBuilder.Append(", ");
				SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.ItemCode.NAME);
				strBuilder.Append(", ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemTable.NAME, PmxItemTable.Columns.CustomItemDescr.NAME);
				strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.Quantity.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.LineNum.NAME);
                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, V33SalesOrderLinesTable.NAME, V33SalesOrderLinesTable.Columns.BladeSwap.NAME);

				strBuilder.Append("\n FROM ");
				strBuilder.Append(SqlCommandHelper.EscapeTableName(SboSalesOrderLineTable.NAME));
				strBuilder.Append("\n LEFT JOIN ");
				strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemTable.NAME));
				strBuilder.Append(" ON ");
				SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.ItemCode.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME);

				strBuilder.Append("\n WHERE ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.DocEntry.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(salesOrderDocEntry, CultureInfo.InvariantCulture));
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.VisOrder.NAME);
                strBuilder.Append(" > ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(salesOrderVisOrderNumber, CultureInfo.InvariantCulture));
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.Quantity.NAME);
                strBuilder.Append(" < ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(Definitions.ZERO, CultureInfo.InvariantCulture));
                strBuilder.Append(" ORDER BY ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboSalesOrderLineTable.NAME, SboSalesOrderLineTable.Columns.VisOrder.NAME);
                strBuilder.Append(" ASC ");

                return strBuilder.ToString();
            }
        }

        #endregion
    }
}
